@@ -17,12 +17,15 @@
  */
 
 #import <AppKit/AppKit.h>
+#import <Foundation/Foundation.h>
 
 #import "AppController.h"
 #import "GNUstep.h"
 
 #import "DictConnection.h"
+#import "LocalDictionary.h"
 #import "NSString+Clickable.h"
+#import "Preferences.h"
 
 NSDictionary* bigHeadlineAttributes;
 NSDictionary* headlineAttributes;
@@ -62,7 +65,7 @@ NSDictionary* normalAttributes;
 }
 
 -(void) writeString: (NSString*) aString
-	 attributes: (NSDictionary*) attributes
+        attributes: (NSDictionary*) attributes
 {
   NSAttributedString* attrStr =
     [[NSAttributedString alloc]
@@ -195,6 +198,8 @@ NSDictionary* normalAttributes;
 @end // AppController (DefinitionWriter)
 
 
+
+
 // FIXME: Just for clean programming style: dealloc is missing,
 //        but I think it's not needed.
 
@@ -203,17 +208,53 @@ NSDictionary* normalAttributes;
 -(id)init
 {
   if (self = [super init]) {
-    dict = [[DictConnection alloc] init];
-    [dict setDefinitionWriter: self];
+    id dict;
     
+    // create mutable dictionaries array
+    dictionaries = [[NSMutableArray alloc] initWithCapacity: 2];
+    
+    NSString* dictStoreFile = [self dictionaryStoreFile];
+    
+    if ([[NSFileManager defaultManager] fileExistsAtPath: dictStoreFile]) {
+        NSArray* plist = [NSArray arrayWithContentsOfFile: [self dictionaryStoreFile]];
+        int i;
+        for (i=0; i<[plist count]; i++) {
+            DictionaryHandle* dict =
+                [DictionaryHandle dictionaryFromPropertyList: [plist objectAtIndex: i]];
+            [dict setDefinitionWriter: self];
+            [dictionaries addObject: dict];
+            NSLog(@" *** Added %@", dict);
+        }
+        NSLog(@" *** result: %@", dictionaries);
+    } else {
+#ifdef PREDEFINED_DICTIONARIES // predefined dictionaries
+        // create local dictionary object
+        dict = [[LocalDictionary alloc] initWithResourceName: @"jargon"];
+        [dict setDefinitionWriter: self];
+        [dictionaries addObject: dict];
+        [dict release];
+#ifdef REMOTE_DICTIONARIES // remote dictionaries
+        // create remote dictionary object
+        dict = [[DictConnection alloc] init];
+        [dict setDefinitionWriter: self];
+        [dictionaries addObject: dict];
+        [dict release];
+#endif // end remote dictionaries block
+#endif // end predefined dictionaries
+    }
+    
+    
+    // create history manager
     historyManager = [[HistoryManager alloc] init];
     [historyManager setDelegate: self];
   }
   
+  
+  // create fonts
   if (bigHeadlineAttributes == nil) {
     bigHeadlineAttributes = 
       [[NSDictionary alloc]
-	initWithObjectsAndKeys:
+	  initWithObjectsAndKeys:
 	  [NSFont titleBarFontOfSize: 16], NSFontAttributeName,
 	nil
        ];
@@ -237,8 +278,11 @@ NSDictionary* normalAttributes;
        ];
   }
   
+  // Notifications --------------
+  NSNotificationCenter* defaultCenter = [NSNotificationCenter defaultCenter];
+  
   // Register in the default notification center to receive link clicked events
-  [[NSNotificationCenter defaultCenter]
+  [defaultCenter
     addObserver: self
     selector: @selector(clickSearchNotification:)
     name: WordClickedNotificationType
@@ -248,6 +292,9 @@ NSDictionary* normalAttributes;
 }
 
 
+
+
+
 // ---- Some methods called by the GUI
 -(void) browseBackClicked: (id)sender {
   [historyManager browseBack];
@@ -257,6 +304,11 @@ NSDictionary* normalAttributes;
   [historyManager browseForward];
 }
 
+-(void) orderFrontPreferencesPanel: (id)sender {
+    [[Preferences shared] setDictionaries: dictionaries];
+    [[Preferences shared] show];
+}
+
 -(void)updateGUI {
   if ([historyManager canBrowseBack]) {
     [browseBackButton setEnabled: YES];
@@ -324,12 +376,15 @@ NSDictionary* normalAttributes;
  */
 -(void)clickSearchNotification: (NSNotification*)aNotification
 {
-  assert(aNotification != nil);
+  NSParameterAssert(aNotification != nil);
   
   // fetch string from notification
   NSString* searchString = [aNotification object];
   
-  assert(searchString != nil);
+  NSAssert1(
+      searchString != nil,
+      @"Search string encapsuled in %@ notification was nil", aNotification
+  );
   
   if ( ![[searchStringControl stringValue] isEqualToString: searchString] ) {
     // invoke search
@@ -337,6 +392,8 @@ NSDictionary* normalAttributes;
   }
 }
 
+
+
 /**
  * Searches for definitions of the word given in the aWord
  * parameter and shows the results in the text view.
@@ -350,36 +407,69 @@ NSDictionary* normalAttributes;
     [searchStringControl setStringValue: aWord];
   }
   
-  NS_DURING
-    {
-      [dict open];
-      [dict sendClientString: @"GNUstep DictionaryReader.app"];
-      [dict definitionFor: aWord];
-      // [dict close]; // FIXME: That crashes!
-      
-      [historyManager browser: self
-		      didBrowseTo: aWord];
-      [self updateGUI];
-      
-      [dictionaryContentWindow orderFront: self];
+  // We need space for new content
+  [self clearResults];
+  
+  // Iterate over all dictionaries, query them!
+  int i;
+  
+  for (i=0; i<[dictionaries count]; i++) {
+    id dict = [dictionaries objectAtIndex: i];
+    
+    if ([dict isActive]) {
+        NS_DURING
+          {
+	    [dict open];
+	    [dict sendClientString: @"GNUstep DictionaryReader.app"];
+	    [dict definitionFor: aWord];
+	    // [dict close];
+          }
+        NS_HANDLER
+          {
+	    NSRunAlertPanel
+	      (
+	       @"Word definition failed.",
+	       [NSString
+	         stringWithFormat:
+	           @"The definition of %@ failed because of this exception:\n%@",
+	         aWord, [localException reason]],
+	       @"Argh", nil, nil);
+          }
+        NS_ENDHANDLER;
     }
-  NS_HANDLER
-    {
-      NSRunAlertPanel
-	(
-	 @"Word definition failed.",
-	 [NSString
-	   stringWithFormat:
-	     @"The definition of %@ failed because of this exception:\n%@",
-	   aWord, [localException reason]],
-	 @"Argh", nil, nil);
-    }
-  NS_ENDHANDLER;
+  }
+  
+  [historyManager browser: self
+		  didBrowseTo: aWord];
+  
+  [self updateGUI];
+  
+  [dictionaryContentWindow orderFront: self];
 }
 
-- (void) applicationDidFinishLaunching: (NSNotification *) theNotification
+
+-(NSString*) dictionaryStoreFile
 {
-  [NSApp setServicesProvider: self];
+#warning FIXME: Ensure that directory exists!
+    return [@"~/GNUstep/Library/DictionaryReader/dictionaries.plist"
+                stringByExpandingTildeInPath];
+}
+
+-(void) applicationWillTerminate: (NSNotification*) theNotification
+{
+    int i;
+    NSMutableArray* mut = [NSMutableArray arrayWithCapacity: [dictionaries count]];
+    
+    for (i=0; i<[dictionaries count]; i++) {
+        [mut addObject: [[dictionaries objectAtIndex: i] shortPropertyList]];
+    }
+    
+    [mut writeToFile: [self dictionaryStoreFile] atomically: YES];
+}
+
+-(void) applicationDidFinishLaunching: (NSNotification*) theNotification
+{
+    [NSApp setServicesProvider: self];
 }
 
 
